import CompuCellSetup
from PySteppables import *
import CompuCell
import sys
# This is to import the deep copy module
from copy import deepcopy
import random
from math import *
from PlayerPython import *
import os
import numpy as np
# import scipy.io as sio

import time
import os.path

# All the parameters of the simulation: Need to write the units

# The parameters that stay fixed
MCSThr = 50  # initial relaxation # of MCS
V0 = 16.0  # tumor cell target volume
S0 = 16.0  # tumor cell target surface
V0_nec = 4.0  # tumor cell volume of necrotic cells
S0_nec = 4.0  # tumor cell surface of necrotic cells
LBD_V0 = 15.0  # lambda value for volume
LBD_S0 = 5.0  # lambda value for surface
LBD_NV0 = 10.0  # lambda value for volume necrotic
LBD_NS0 = 5.0  # lambda value for surface necrotic
EV0 = 40.0 * V0  # tumor cell target volume
ES0 = 40.0 * S0  # tumor cell target surface
LBD_EV0 = 40.0 * LBD_V0  # lambda value for volume
LBD_ES0 = 40.0 * LBD_S0  # lambda value for surface
kgts = 4.0  # factor multiplying: surface = ktgs*sqrt(volume)
Pvolmaxmit = 2.0 * V0  # tumor cell mitosis volume
Svolmaxmit = 2.0 * V0  # stem cell mitosis volume
a = 1  # the Hill coefficient

# ***** The parameters that change *******
# The constant values: For cell growth
incvol = 13.0  # growth rate of tumor cells in pixels/MCS
decvol = 0.025  # decreasing rate of tumor cells in pixels/MCS
PGrThr = 0.1  # concetration above which proliferating cell grows: 10% of ATP = 10% of 5mM
SGrThr = 0.1  # concetration above which stem cell grows: 10% of ATP# percentage of LAC from environment
StressIncrement = 0.95  # Stress increment due to neighbors

# Constant for the starvation function
PUgMax = 0.6  # Proliferating cell -  maximum uptaken rate
QUgMax = 0.75 * PUgMax  # Quiescent cell -  maximum uptaken rate
atpK = 0.32  # 5mM ;Michaelis-Menten constant for ATP concentration
atpD = 0.02698  # ATP concentration below which damage accumulates --- 0.05mM --> 0.0032 fmol/vox

# Thresholds for the starvation 
Total_time = 190  # 12 hours in MCS
Neg_concATP = 0.0011259  # Zero concentration for 12 hours
PNeThr = Total_time * (PUgMax * (
        Neg_concATP ** a / (Neg_concATP ** a + atpK ** a)))  # .total damage at which:proliferating -> necrotic :

QCNeThr = 2 * PNeThr  # total damage at which: quiescent -> necrotic : intially 2*102.0
SNeThr = 4 * PNeThr  # total damage at which: stem -> necrotic
QSNeThr = 8 * PNeThr  # total damage at which: quiescent -> necrotic

# Thresholds for the happiness
Pos_concATP = 1.4007  # Total 5 mM concentration for 12 hours
QCPThr = Total_time * (QUgMax * (
        Pos_concATP ** a / (Pos_concATP ** a + atpK ** a)))  # total happiness at which: quiescent -> proliferating
QSSThr = QCPThr  # total happiness at which: QS -> S

# Next, are the thresholds for the acidosis set in
PLUgMax = 15.0  # Proliferating cell -  maximum uptaken rate = 10e-17 mol/cell/sec --> 2.25e-15 fmol/vox/MCS
QLUgMax = 0.75 * PLUgMax  # Quiescent cell -  maximum uptaken rate
lacK = 0.064  # Michaelis-Menten constant for ATP concentration = 0.04 mM --> 2.56e-18 mol/vox --> 0.00256 fmol/vox
LacDeath = 1.8754  # 50mM Threshold at which the necrosis sets in due to acidity from Lactate concentration

# Only necrosis thresholds for Lactate concentrations
Pos_concLac = 2.5449  # 60 mM Lactate causes death
Total_time_lac = 336
PNe_AThr = Total_time_lac * (PLUgMax * (
        Pos_concLac ** a / (Pos_concLac ** a + lacK ** a)))  # .total damage at which:proliferating -> necrotic :
QCNe_AThr = 2 * PNe_AThr  # total damage at which: quiescent -> necrotic
SNe_AThr = 4 * PNe_AThr  # total damage at which: stem -> necrotic :
QSNe_AThr = 8 * PNe_AThr  # total damage at which: quiescent -> necrotic

# Other parameters
maxdiv = 9  # =8 maximum number of cell divisions
probstem = 0.2  # =0.20 probability of new stem cell after division
StressThr = 50  # Cell Stress threshold time as a counter for MCS to vary from 20 - 100
K = 6
Neigh = 6


# Get the centre of mass coordinates
def getCellCOMPoint3D(cell):
    pt = CompuCell.Point3D()
    pt.x = int(round(cell.xCOM))
    pt.y = int(round(cell.yCOM))
    pt.z = int(round(cell.zCOM))

    return pt


from PySteppables import *
import CompuCell
import sys

from PlayerPython import *
import CompuCellSetup
from math import *


# This is the epithelial Sheet Steppable.
class EpithelialSheetSteppable(SteppableBasePy):
    def __init__(self, _simulator, _frequency=1):
        SteppableBasePy.__init__(self, _simulator, _frequency)

    def step(self, mcs):
        # Turn on constriction

        for cell in self.cellListByType(self.BASAL, self.ECM):  # list of  cell types (capitalized)
            if 0 < cell.xCOM < 400:
                # Make sure FocalPointPlacticity plugin is loaded
                # visiting all focal links cell 'cell' has with other cells
                for fppd in self.getFocalPointPlasticityDataList(cell):

                    lambdaD = 20.0 * fppd.lambdaDistance
                    # YOU DO NOT WANT TO KEEP MULTIPLYING lambdaD BECAUSE IT EVENTUALLY BECOMES INF(inity) ...
                    # ALSO MAKE SURE THAT THE LOGIC IN THIS PART OF CODE IS WHAT YOU REALY WANT
                    if lambdaD > 1e8:
                        continue
                    targetD = fppd.targetDistance
                    # print 'lambdaD=', lambdaD
                    # arguments are (cell1,cell2,lambda,targetDistance,maxDistance)
                    self.focalPointPlasticityPlugin.setFocalPointPlasticityParameters(cell, fppd.neighborAddress,
                                                                                      lambdaD, targetD,
                                                                                      5.0)






from PySteppables import *
import CompuCell
import sys

from PlayerPython import *
import CompuCellSetup
from math import *


# This is the SBML solver Steppable: Carried out at every time step
class SBMLSolverSteppable(SteppableBasePy):
    def __init__(self, _simulator, _frequency=1):
        SteppableBasePy.__init__(self, _simulator, _frequency)

    def start(self):
        options = {'relative': 1e-09, 'absolute': 1e-12}
        self.setSBMLGlobalOptions(options)
        modelFile = './Simulation/S3_diffusion.xml'
        # StepSize is the time Unit : time in SBML(integration time Step)
        # this is the total time fow which the SBML runs
        stepSize = 6
        # Add the SBML model to each cell
        # iterating over all cells in simulation
        self.addSBMLToCellTypes(_modelFile=modelFile, _modelName='PC_ODE',
                                _types=[self.PCANCER, self.QCANCER, self.PSTEM, self.QSTEM],
                                _stepSize=stepSize)

        # Initiate the cell Dictionary at the beginning of the steppable
        for cell in self.cellListByType(self.PCANCER, self.QCANCER, self.PSTEM, self.QSTEM):
            cellDict = CompuCell.getPyAttrib(cell)
            cellDict["Glu_in"] = 0
            cellDict["Glutamine_in"] = 0
            cellDict["ATP"] = 0
            cellDict["exLactate"] = 0
            cellDict["exLactate_now"] = 0
            cellDict["Lactate_in"] = 0

    def step(self, mcs):
        # Chemical Field Manipulation -> Get field reference    
        glucoseField = self.getConcentrationField('Glucose')
        glutamineField = self.getConcentrationField('Glutamine')
        oxygenField = self.getConcentrationField('Oxygen')
        LactateField = self.getConcentrationField('Lactate')
        # iterating over cells of type PC,PS,QS,QC        
        for cell in self.cellListByType(self.PCANCER, self.QCANCER, self.PSTEM, self.QSTEM):
            cellDict = CompuCell.getPyAttrib(cell)
            pt = getCellCOMPoint3D(cell)
            # Get the value of the nutrient concentration at the centre of mass
            # This is a constant concentration of how much ever is present in the environment 
            concg = glucoseField.get(pt)
            concgl = glutamineField.get(pt)
            concox = oxygenField.get(pt)
            conclac = LactateField.get(pt)
            # converting them to mM
            Glu_out = (5 / 0.32) * (concg)
            Glutamine_out = (5 / 0.32) * (concgl)
            O2e = (5 / 0.32) * (concox)
            Lac_out = (5 / 0.32) * (conclac)

            Glu_out = max(0.0001, Glu_out)
            O2e = max(0.0001, O2e)
            Glutamine_out = max(0.0001, Glutamine_out)
            if Glu_out < 0.0005:
                Lac_out = max(0.0001, Lac_out)  # Uptake lactate only when glucose is zero
            else:
                Lac_out = 0.0001  # Lactate taken up is zero present in the cells

            # Set the SBML values of the nutrients from the extracellular field
            self.setSBMLValue(_modelName='PC_ODE', _valueName='Glu_out', _value=Glu_out,
                              _cell=cell)  # This is glucose out
            self.setSBMLValue(_modelName='PC_ODE', _valueName='Glutamine_out', _value=Glutamine_out,
                              _cell=cell)  # This is glutamine out
            self.setSBMLValue(_modelName='PC_ODE', _valueName='O2e', _value=O2e,
                              _cell=cell)  # This is external oxygen
            self.setSBMLValue(_modelName='PC_ODE', _valueName='Lac_out', _value=Lac_out,
                              _cell=cell)  # This is Lactate out
            # MCS is one step ahead of SBML, this makes SBML step only when MCS > 1
            if mcs > 1:
                self.timestepSBML()
            # Keep track of the ATP produced at each MCS for each cell type
            state = self.getSBMLState(_modelName='PC_ODE', _cell=cell)  # returns dictionary of values
            # nATP = self.getSBMLValue(_modelName='PC_ODE', _valueName='ATP', _cell=cell)  
            cellDict["Glu_in"] = state['Glu_in']
            cellDict["Glutamine_in"] = state['Glutamine_in']
            cellDict["ATP"] = state['ATP']
            cellDict["exLactate"] = state['Lac_out']
            cellDict["Lactate_in"] = state['Lac_in']
            # Setting the lactate field to the current production from cells
            LactateField.set(pt, (0.32 / 5) * cellDict["exLactate"])
            cellDict["exLactate_now"] = LactateField.get(pt)


# This steppable shows the growth of the proliferating cells 
# and the shrinkage of the necrotic cells
class TumorGrowthSteppable(SteppableBasePy):

    def __init__(self, _simulator, _frequency):
        SteppableBasePy.__init__(self, _simulator, _frequency)

    def start(self):

        # This is the cell growth for the tumor cells
        for cell in self.cellListByType(self.PCANCER, self.QCANCER, self.PSTEM, self.QSTEM):
            cell.targetVolume = V0
            cell.lambdaVolume = LBD_V0
            cell.targetSurface = S0
            cell.lambdaSurface = LBD_S0

            cellDict = CompuCell.getPyAttrib(cell)

            cellDict["Counter"] = 0
            cellDict["Health"] = 0
            cellDict["Starvation"] = 0
            cellDict["Acidity"] = 0
            cellDict["Stress"] = 0
        for cell in self.cellListByType(self.NECROTIC):
            cell.targetVolume = V0_nec
            cell.lambdaVolume = LBD_NV0
            cell.targetSurface = S0_nec
            cell.lambdaSurface = LBD_NS0
        # The ECM and Basal cells have different target and lambda volume
        for cell in self.cellListByType(self.BASAL, self.ECM):
            cell.targetVolume = EV0
            cell.lambdaVolume = LBD_EV0
            cell.targetSurface = ES0
            cell.lambdaSurface = LBD_ES0

    def step(self, mcs):
        # runs every 10 MCS
        LactateField = self.getConcentrationField('Lactate')
        # Every 50 MCS
        if mcs % 50 == 0:
            for cell in self.cellList:
                LactateField = self.getConcentrationField('Lactate')
                # Get the attributes of the cell
                cellDict = CompuCell.getPyAttrib(cell)
                if cell.type == self.PCANCER:
                    conc_gr = (0.32 / 5) * (cellDict["Glu_in"] + cellDict["Glutamine_in"] + cellDict["ATP"])
                    cell.targetVolume += incvol * max(0, conc_gr - PGrThr)
                    cell.targetSurface = kgts * sqrt(cell.targetVolume)
                # Stem Cells
                if cell.type == self.PSTEM:
                    conc_gr = (0.32 / 5) * (cellDict["Glu_in"] + cellDict["Glutamine_in"] + cellDict["ATP"])
                    cell.targetVolume += incvol * max(0, conc_gr - SGrThr)
                    cell.targetSurface = kgts * sqrt(cell.targetVolume)
                # If the cells have become necrotic, they shrink
                # They cannot change their state irrespective of the amount of glucose present
                if cell.type == self.NECROTIC:
                    cellNeighborList = self.getCellNeighborDataList(cell)
                    neighbor_count_by_type = cellNeighborList.neighborCountByType()
                    # only the cells at the periphery shrink and die
                    if neighbor_count_by_type[0] >= 1:
                        cell.targetVolume -= min(decvol, cell.targetVolume)

                    else:  # cells in the core degrade and thier target volume is 1/4th of the original
                        cell.targetVolume -= 0

                    cell.targetSurface = kgts * sqrt(cell.targetVolume)


class CellStateTransition(SteppableBasePy):
    def __init__(self, _simulator, _frequency=1):
        SteppableBasePy.__init__(self, _simulator, _frequency)

    # Not there in the paper code, added this to be consistent with the previous steppables
    def start(self):

        for cell in self.cellListByType(self.PCANCER, self.QCANCER, self.PSTEM, self.QSTEM):
            cellDict = CompuCell.getPyAttrib(cell)

    def step(self, mcs):
        # Every 50 MCS
        if mcs % 50 == 0:
            LactateField = self.getConcentrationField('Lactate')
            for cell in self.cellListByType(self.PCANCER, self.QCANCER, self.PSTEM, self.QSTEM):
                cellDict = CompuCell.getPyAttrib(cell)
                XLac = cellDict["Lactate_in"]
                XLacout = cellDict["exLactate"]
                pt = getCellCOMPoint3D(cell)
                # Proliferating -> Necrotic
                # If the cell has aquired enough damage it will turn into necrotic
                # Once necrotic, its health becomes zero
                # Cells transfer its complete lactate out to the microenvironment

                # Introduce stress induced transition
                if (
                        cell.type == self.QSTEM or cell.type == self.QCANCER or cell.type == self.PSTEM or cell.type == self.PCANCER):
                    cellNeighborList = self.getCellNeighborDataList(cell)
                    neighbor_count_by_type = cellNeighborList.neighborCountByType()
                    # Count the total number of cell numbers:
                    allPC = neighbor_count_by_type[1L]
                    allQC = neighbor_count_by_type[2L]
                    allPS = neighbor_count_by_type[3L]
                    allQS = neighbor_count_by_type[4L]
                    allNec = neighbor_count_by_type[5L]
                    All_neighbor = allPC + allQC + allPS + allQS + allNec
                    if neighbor_count_by_type[0] == 0 and All_neighbor >= int(Neigh):
                        cellDict["Stress"] += StressIncrement

                if cell.type == self.PCANCER:
                    if (cellDict["Starv"] > PNeThr or cellDict["Acidity"] > PNe_AThr or cellDict["Stress"] > StressThr):
                        LactateField.set(pt, (0.32 / 5) * (XLac + XLacout))
                        cellDict["exLactate_now"] = LactateField.get(pt)
                        cell.type = self.NECROTIC
                        cellDict["Health"] = 0

                # Stem -> Necrotic : Transfer all lactate before changing into necrotic
                if cell.type == self.PSTEM:
                    if (cellDict["Starv"] > SNeThr or cellDict["Acidity"] > SNe_AThr or cellDict["Stress"] > StressThr):
                        LactateField.set(pt, (0.32 / 5) * (XLac + XLacout))
                        cellDict["exLactate_now"] = LactateField.get(pt)
                        cell.type = self.NECROTIC
                        cellDict["Health"] = 0

                # QuiescentCancer  -> Necrotic
                # QuiescentCancer -> PCANCER
                if cell.type == self.QCANCER:
                    if (cellDict["Starv"] > QCNeThr or cellDict["Acidity"] > QCNe_AThr or cellDict[
                        "Stress"] > StressThr):
                        LactateField.set(pt, (0.32 / 5) * (XLac + XLacout))
                        cellDict["exLactate_now"] = LactateField.get(pt)
                        cell.type = self.NECROTIC
                        cellDict["Health"] = 0
                    # Has to gain enough health to become proliferating Cancer
                    if cellDict["Health"] > QCPThr:
                        cell.type = self.PCANCER
                        cellDict["Health"] = 0

                # Quiescent Stem -> Necrotic
                # Quiescent Stem -> Stem
                if cell.type == self.QSTEM:
                    if (cellDict["Starv"] > QSNeThr or cellDict["Acidity"] > QSNe_AThr or cellDict[
                        "Stress"] > StressThr):
                        LactateField.set(pt, (0.32 / 5) * (XLac + XLacout))
                        cellDict["exLactate_now"] = LactateField.get(pt)
                        cell.type = self.NECROTIC
                        cellDict["Health"] = 0
                    # Gains health to become Proliferating Stem
                    if cellDict["Health"] > QSSThr:
                        cell.type = self.PSTEM
                        cellDict["Health"] = 0


from PySteppables import *
import CompuCell
import sys

from PlayerPython import *
import CompuCellSetup
from math import *


class StravHealthCalculator(SteppableBasePy):
    def __init__(self, _simulator, _frequency=1):
        SteppableBasePy.__init__(self, _simulator, _frequency)

    # Setting the health and starvation counters to zero    
    def start(self):

        for cell in self.cellListByType(self.PCANCER, self.QCANCER, self.PSTEM, self.QSTEM, self.NECROTIC):
            cellDict = CompuCell.getPyAttrib(cell)
            # Since we calculate the starvation and health factors we start with zero
            cellDict["Starv"] = 0
            cellDict["Health"] = 0
            cellDict["Acidity"] = 0
            cellDict["Stress"] = 0

    # Define the Michelis menten
    def MM(self, x, m, k, coeff):

        return (m * (x ** coeff / (x ** coeff + k ** coeff)))

    def step(self, mcs):
        # Every 50 MCS
        if mcs % 50 == 0:
            for cell in self.cellListByType(self.PCANCER, self.QCANCER, self.PSTEM, self.QSTEM):
                cellDict = CompuCell.getPyAttrib(cell)
                # For the cancer (proliferating cells)
                if (cell.type == self.PCANCER or cell.type == self.PSTEM):
                    conc_atp = (0.32 / 5) * cellDict["ATP"]
                    conc_gr = (0.32 / 5) * (cellDict["Glu_in"] + cellDict["Glutamine_in"] + cellDict["ATP"])
                    conc_lacout = (0.32 / 5) * cellDict["exLactate_now"]
                    if conc_atp < atpD:  # If Conc is less than 0.032 they increase the starvation factor
                        # MM(self,x,m,k), increase the starvation coefficient
                        cellDict["Starv"] += abs(self.MM(conc_atp, PUgMax, atpK, a) - self.MM(atpD, PUgMax, atpK, a))
                    elif conc_lacout > LacDeath:
                        cellDict["Acidity"] += abs(
                            self.MM(conc_lacout, PLUgMax, lacK, a) - self.MM(LacDeath, PLUgMax, lacK, a))
                    else:  # If the conc is more than 0.032 the cells aquire health, increase the health coefficient
                        cellDict["Health"] += exp(K) * (self.MM(conc_gr, PUgMax, atpK, a))

                if (cell.type == self.QCANCER or cell.type == self.QSTEM):
                    conc_atp = (0.32 / 5) * cellDict["ATP"]
                    conc_gr = (0.32 / 5) * (cellDict["Glu_in"] + cellDict["Glutamine_in"] + cellDict["ATP"])
                    conc_lacout = (0.32 / 5) * cellDict["exLactate_now"]
                    if conc_atp < atpD:
                        cellDict["Starv"] += abs(self.MM(conc_atp, QUgMax, atpK, a) - self.MM(atpD, QUgMax, atpK, a))
                    elif conc_lacout > LacDeath:
                        cellDict["Acidity"] += abs(
                            self.MM(conc_lacout, QLUgMax, lacK, a) - self.MM(LacDeath, QLUgMax, lacK, a))
                    else:
                        cellDict["Health"] += exp(K) * (self.MM(conc_gr, QUgMax, atpK, a))

    def finish(self):
        # this function may be called at the end of simulation - used very infrequently though
        return


from PySteppables import *
from PySteppablesExamples import MitosisSteppableBase
import CompuCell
import sys

from PlayerPython import *
from math import *


class MitosisSteppable(MitosisSteppableBase):
    def __init__(self, _simulator, _frequency=1):
        MitosisSteppableBase.__init__(self, _simulator, _frequency)

    def start(self):

        # iterating over all cells in simulation
        for cell in self.cellListByType(self.PCANCER, self.QCANCER, self.PSTEM, self.QSTEM):
            # Set the counter for division to be zero
            cellDict = CompuCell.getPyAttrib(cell)
            cellDict["Counter"] = 0  # Number of divisions

    def step(self, mcs):

        cells_to_divide = []
        if mcs % 50 == 0:
            for cell in self.cellList:
                # When a cell reaches its doubling volume, it divides
                # It gains volume by growing in the presence of enough glucose
                if ((cell.type == self.PCANCER or cell.type == self.QCANCER) and cell.volume > Pvolmaxmit) or (
                        (cell.type == self.PSTEM or cell.type == self.QSTEM) and cell.volume > Svolmaxmit):
                    cells_to_divide.append(cell)

            for cell in cells_to_divide:
                # cells divide along a random axis
                self.divideCellRandomOrientation(cell)

    def updateAttributes(self):

        # Mitosis resets the parent and daughter generalised cell's
        # Target volume and surface to their reference values
        # and their counters, health and damage to zero
        LactateField = self.getConcentrationField('Lactate')
        parentCell = self.mitosisSteppable.parentCell
        childCell = self.mitosisSteppable.childCell
        parentCell.targetVolume = V0
        childCell.targetVolume = V0
        parentCell.targetSurface = kgts * sqrt(parentCell.targetVolume)
        childCell.targetSurface = kgts * sqrt(childCell.targetVolume)
        parentCell.lambdaVolume = LBD_V0
        parentCell.lambdaSurface = LBD_S0
        childCell.lambdaVolume = LBD_V0
        childCell.lambdaSurface = LBD_S0
        State_parentcell = self.getSBMLState(_modelName='PC_ODE', _cell=parentCell)  # Get all the states
        self.copySBMLs(_fromCell=parentCell, _toCell=childCell)  # Copy the SBML to the child Cell
        self.setSBMLState(_modelName='PC_ODE', _cell=childCell, _state=State_parentcell)
        pt_p = getCellCOMPoint3D(parentCell)
        pt_c = getCellCOMPoint3D(childCell)
        # Get the attibutes of each type of cell
        parentCellDict = CompuCell.getPyAttrib(parentCell)
        childCellDict = CompuCell.getPyAttrib(childCell)
        statep = self.getSBMLState(_modelName='PC_ODE', _cell=parentCell)  # returns dictionary of values
        # pATP = self.getSBMLValue(_modelName='PC_ODE', _valueName='ATP', _cell=parentCell)
        parentCellDict["ATP"] = statep['ATP']
        parentCellDict["Glu_in"] = statep['Glu_in']
        parentCellDict["Glutamine_in"] = statep['Glutamine_in']
        parentCellDict["exLactate"] = statep['Lac_out']
        parentCellDict["exLactate_now"] = LactateField.get(pt_p)
        parentCellDict["Lactate_in"] = statep['Lac_in']

        statec = self.getSBMLState(_modelName='PC_ODE', _cell=childCell)
        # cATP = self.getSBMLValue(_modelName='PC_ODE', _valueName='ATP', _cell=childCell)
        childCellDict["ATP"] = statec['ATP']
        childCellDict["Glu_in"] = statec['Glu_in']
        childCellDict["Glutamine_in"] = statec['Glutamine_in']
        childCellDict["exLactate"] = statec['Lac_out']
        childCellDict["Lactate_in"] = statec['Lac_in']
        childCellDict["exLactate_now"] = LactateField.get(pt_c)

        if parentCell.type == self.PCANCER:
            # Pick a random number that this cell
            # can undergo divisions 
            # After which it becomes necrotic
            temp = random.gauss(maxdiv, 2)
            if parentCellDict["Counter"] <= temp:
                # PC becomes QC after division
                parentCell.type = self.QCANCER  # both cells are Q after mitosis
                childCell.type = self.QCANCER
                parentCellDict["Counter"] += 1

            childCellDict["Counter"] = deepcopy(parentCellDict["Counter"])

            # If it has already undergone the maximum number of division
            # The parent and daughter cells become necrotic
            if parentCellDict["Counter"] > temp:
                parentCell.type = self.NECROTIC
                childCell.type = self.NECROTIC

        if parentCell.type == self.PSTEM or parentCell.type == self.QSTEM:  # Stem cell
            parentCellDict["Counter"] += 1
            parentCell.type = self.QSTEM  # One is QS and
            childCell.type = self.QCANCER  # the other is QC

            if random.random() <= probstem:  # But there is a chance for a 2nd QS
                childCell.type = self.QSTEM
            childCellDict["Counter"] = 0

        parentCellDict["Starv"] = 0
        childCellDict["Starv"] = 0
        parentCellDict["Acidity"] = 0
        childCellDict["Acidity"] = 0
        parentCellDict["Health"] = 0
        childCellDict["Health"] = 0
        parentCellDict["Stress"] = 0
        childCellDict["Stress"] = 0


from PySteppables import *
import CompuCell
import sys

from PlayerPython import *
import CompuCellSetup
from math import *


class DataOutput(SteppableBasePy):
    def __init__(self, _simulator, _frequency=10):
        SteppableBasePy.__init__(self, _simulator, _frequency)
        self.outStr = ''

    def start(self):
        self.outfile1, outfile1_name = self.openFileInSimulationOutputDirectory("Data_necrotic.dat", "w")
        self.outfile2, outfile2_name = self.openFileInSimulationOutputDirectory("Data_PC.dat", "w")
        self.outfile3, outfile3_name = self.openFileInSimulationOutputDirectory("Data_PS.dat", "w")
        self.outfile4, outfile4_name = self.openFileInSimulationOutputDirectory("Data_QC.dat", "w")
        self.outfile5, outfile5_name = self.openFileInSimulationOutputDirectory("Data_QS.dat", "w")
        self.outfile6, outfile6_name = self.openFileInSimulationOutputDirectory("NumberOfCells.dat", "a")

    def step(self, mcs):
        # Every 200th MCS = 1200 min = 0.84 day
        if mcs % 300 == 0:

            for cell in self.cellListByType(self.NECROTIC):
                # data_list1=[cell.id,cell.volume,cell.surface,cell.xCOM,cell.yCOM]
                data_list1 = str(cell.id) + "\t" + str(cell.volume) + "\t"
                data_list1 += str(cell.surface) + "\t" + str(cell.xCOM) + "\t" + str(cell.yCOM) + "\t"
                data_list1 += "\n"
                self.outfile1.write(data_list1)

            for cell in self.cellListByType(self.PCANCER):
                state = self.getSBMLState(_modelName='PC_ODE', _cell=cell)
                data_list2 = str(cell.id) + "\t" + str(cell.volume) + "\t" + str(cell.surface) + "\t" + str(
                    cell.xCOM) + "\t" + str(cell.yCOM) + "\t"
                data_list2 += str(max(0, state['Glu_out'])) + "\t" + str(max(0, state['Glutamine_out'])) + "\t"
                data_list2 += str(max(0, state['Lac_out'])) + "\t" + str(max(0, state['O2e'])) + "\t"
                data_list2 += str(max(0, state['ADP'])) + "\t" + str(max(0, state['AKG'])) + "\t"
                data_list2 += str(max(0, state['ASP'])) + "\t" + str(max(0, state['ATP'])) + "\t"
                data_list2 += str(max(0, state['CIT'])) + "\t" + str(max(0, state['DHAP'])) + "\t"
                data_list2 += str(max(0, state['E4P'])) + "\t" + str(max(0, state['F6P'])) + "\t"
                data_list2 += str(max(0, state['G3P'])) + "\t" + str(max(0, state['Glu_in'])) + "\t"
                data_list2 += str(max(0, state['Glutamate'])) + "\t" + str(max(0, state['Glutamine_in'])) + "\t"
                data_list2 += str(max(0, state['GSH'])) + "\t" + str(max(0, state['Lac_in'])) + "\t"
                data_list2 += str(max(0, state['mAKG'])) + "\t" + str(max(0, state['MAL'])) + "\t"
                data_list2 += str(max(0, state['mASP'])) + "\t" + str(max(0, state['mCIT'])) + "\t"
                data_list2 += str(max(0, state['mFUM'])) + "\t" + str(max(0, state['mGlutamate'])) + "\t"
                data_list2 += str(max(0, state['mICIT'])) + "\t" + str(max(0, state['mMAL'])) + "\t"
                data_list2 += str(max(0, state['mOAA'])) + "\t" + str(max(0, state['mSUC'])) + "\t"
                data_list2 += str(max(0, state['NADH'])) + "\t" + str(max(0, state['NADPH'])) + "\t"
                data_list2 += str(max(0, state['O2i'])) + "\t" + str(max(0, state['OAA'])) + "\t"
                data_list2 += str(max(0, state['Glutamate'])) + "\t" + str(max(0, state['Glutamine_in'])) + "\t"
                data_list2 += str(max(0, state['PEP'])) + "\t" + str(max(0, state['PYR'])) + "\t"
                data_list2 += str(max(0, state['Pyr'])) + "\t" + str(max(0, state['R5P'])) + "\t"
                data_list2 += str(max(0, state['S7P'])) + "\t" + str(max(0, state['sixPG'])) + "\t" + str(
                    max(0, state['threePG'])) + "\t"
                data_list2 += "\n"
                self.outfile2.write(data_list2)
            for cell in self.cellListByType(self.PSTEM):
                state = self.getSBMLState(_modelName='PC_ODE', _cell=cell)
                data_list3 = str(cell.id) + "\t" + str(cell.volume) + "\t" + str(cell.surface) + "\t" + str(
                    cell.xCOM) + "\t" + str(cell.yCOM) + "\t"
                data_list3 += str(max(0, state['Glu_out'])) + "\t" + str(max(0, state['Glutamine_out'])) + "\t"
                data_list3 += str(max(0, state['Lac_out'])) + "\t" + str(max(0, state['O2e'])) + "\t"
                data_list3 += str(max(0, state['ADP'])) + "\t" + str(max(0, state['AKG'])) + "\t"
                data_list3 += str(max(0, state['ASP'])) + "\t" + str(max(0, state['ATP'])) + "\t"
                data_list3 += str(max(0, state['CIT'])) + "\t" + str(max(0, state['DHAP'])) + "\t"
                data_list3 += str(max(0, state['E4P'])) + "\t" + str(max(0, state['F6P'])) + "\t"
                data_list3 += str(max(0, state['G3P'])) + "\t" + str(max(0, state['Glu_in'])) + "\t"
                data_list3 += str(max(0, state['Glutamate'])) + "\t" + str(max(0, state['Glutamine_in'])) + "\t"
                data_list3 += str(max(0, state['GSH'])) + "\t" + str(max(0, state['Lac_in'])) + "\t"
                data_list3 += str(max(0, state['mAKG'])) + "\t" + str(max(0, state['MAL'])) + "\t"
                data_list3 += str(max(0, state['mASP'])) + "\t" + str(max(0, state['mCIT'])) + "\t"
                data_list3 += str(max(0, state['mFUM'])) + "\t" + str(max(0, state['mGlutamate'])) + "\t"
                data_list3 += str(max(0, state['mICIT'])) + "\t" + str(max(0, state['mMAL'])) + "\t"
                data_list3 += str(max(0, state['mOAA'])) + "\t" + str(max(0, state['mSUC'])) + "\t"
                data_list3 += str(max(0, state['NADH'])) + "\t" + str(max(0, state['NADPH'])) + "\t"
                data_list3 += str(max(0, state['O2i'])) + "\t" + str(max(0, state['OAA'])) + "\t"
                data_list3 += str(max(0, state['Glutamate'])) + "\t" + str(max(0, state['Glutamine_in'])) + "\t"
                data_list3 += str(max(0, state['PEP'])) + "\t" + str(max(0, state['PYR'])) + "\t"
                data_list3 += str(max(0, state['Pyr'])) + "\t" + str(max(0, state['R5P'])) + "\t"
                data_list3 += str(max(0, state['S7P'])) + "\t" + str(max(0, state['sixPG'])) + "\t" + str(
                    max(0, state['threePG'])) + "\t"
                data_list3 += "\n"
                self.outfile3.write(data_list3)
            for cell in self.cellListByType(self.QCANCER):
                state = self.getSBMLState(_modelName='PC_ODE', _cell=cell)
                # data_list4=[cell.id,cell.volume,cell.surface,cell.xCOM,cell.yCOM,state['Glu_out'],state['Glutamine_out'],state['Lac_out'],state['O2e'],state['ADP'],state['AKG'],state['ASP'],state['ATP'],state['CIT'],state['DHAP'],state['E4P'],state['F6P'],state['G3P'],state['Glu_in'],state['Glutamate'],state['Glutamine_in'],state['GSH'],state['Lac_in'],state['mAKG'],state['MAL'],state['mASP'],state['mCIT'],state['mFUM'],state['mGlutamate'],state['mICIT'],state['mMAL'],state['mOAA'],state['mSUC'],state['NADH'],state['NADPH'],state['O2i'],state['OAA'],state['PEP'],state['PYR'],state['Pyr'],state['R5P'],state['S7P'],state['sixPG'],state['threePG']]
                data_list4 = str(cell.id) + "\t" + str(cell.volume) + "\t" + str(cell.surface) + "\t" + str(
                    cell.xCOM) + "\t" + str(cell.yCOM) + "\t"
                data_list4 += str(max(0, state['Glu_out'])) + "\t" + str(max(0, state['Glutamine_out'])) + "\t"
                data_list4 += str(max(0, state['Lac_out'])) + "\t" + str(max(0, state['O2e'])) + "\t"
                data_list4 += str(max(0, state['ADP'])) + "\t" + str(max(0, state['AKG'])) + "\t"
                data_list4 += str(max(0, state['ASP'])) + "\t" + str(max(0, state['ATP'])) + "\t"
                data_list4 += str(max(0, state['CIT'])) + "\t" + str(max(0, state['DHAP'])) + "\t"
                data_list4 += str(max(0, state['E4P'])) + "\t" + str(max(0, state['F6P'])) + "\t"
                data_list4 += str(max(0, state['G3P'])) + "\t" + str(max(0, state['Glu_in'])) + "\t"
                data_list4 += str(max(0, state['Glutamate'])) + "\t" + str(max(0, state['Glutamine_in'])) + "\t"
                data_list4 += str(max(0, state['GSH'])) + "\t" + str(max(0, state['Lac_in'])) + "\t"
                data_list4 += str(max(0, state['mAKG'])) + "\t" + str(max(0, state['MAL'])) + "\t"
                data_list4 += str(max(0, state['mASP'])) + "\t" + str(max(0, state['mCIT'])) + "\t"
                data_list4 += str(max(0, state['mFUM'])) + "\t" + str(max(0, state['mGlutamate'])) + "\t"
                data_list4 += str(max(0, state['mICIT'])) + "\t" + str(max(0, state['mMAL'])) + "\t"
                data_list4 += str(max(0, state['mOAA'])) + "\t" + str(max(0, state['mSUC'])) + "\t"
                data_list4 += str(max(0, state['NADH'])) + "\t" + str(max(0, state['NADPH'])) + "\t"
                data_list4 += str(max(0, state['O2i'])) + "\t" + str(max(0, state['OAA'])) + "\t"
                data_list4 += str(max(0, state['Glutamate'])) + "\t" + str(max(0, state['Glutamine_in'])) + "\t"
                data_list4 += str(max(0, state['PEP'])) + "\t" + str(max(0, state['PYR'])) + "\t"
                data_list4 += str(max(0, state['Pyr'])) + "\t" + str(max(0, state['R5P'])) + "\t"
                data_list4 += str(max(0, state['S7P'])) + "\t" + str(max(0, state['sixPG'])) + "\t" + str(
                    max(0, state['threePG'])) + "\t"
                data_list4 += "\n"
                self.outfile4.write(data_list4)

            for cell in self.cellListByType(self.QSTEM):
                state = self.getSBMLState(_modelName='PC_ODE', _cell=cell)
                # data_list5=[cell.id,cell.volume,cell.surface,cell.xCOM,cell.yCOM,state['Glu_out'],state['Glutamine_out'],state['Lac_out'],state['O2e'],state['ADP'],state['AKG'],state['ASP'],state['ATP'],state['CIT'],state['DHAP'],state['E4P'],state['F6P'],state['G3P'],state['Glu_in'],state['Glutamate'],state['Glutamine_in'],state['GSH'],state['Lac_in'],state['mAKG'],state['MAL'],state['mASP'],state['mCIT'],state['mFUM'],state['mGlutamate'],state['mICIT'],state['mMAL'],state['mOAA'],state['mSUC'],state['NADH'],state['NADPH'],state['O2i'],state['OAA'],state['PEP'],state['PYR'],state['Pyr'],state['R5P'],state['S7P'],state['sixPG'],state['threePG']]
                data_list5 = str(cell.id) + "\t" + str(cell.volume) + "\t" + str(cell.surface) + "\t" + str(
                    cell.xCOM) + "\t" + str(cell.yCOM) + "\t"
                data_list5 += str(max(0, state['Glu_out'])) + "\t" + str(max(0, state['Glutamine_out'])) + "\t"
                data_list5 += str(max(0, state['Lac_out'])) + "\t" + str(max(0, state['O2e'])) + "\t"
                data_list5 += str(max(0, state['ADP'])) + "\t" + str(max(0, state['AKG'])) + "\t"
                data_list5 += str(max(0, state['ASP'])) + "\t" + str(max(0, state['ATP'])) + "\t"
                data_list5 += str(max(0, state['CIT'])) + "\t" + str(max(0, state['DHAP'])) + "\t"
                data_list5 += str(max(0, state['E4P'])) + "\t" + str(max(0, state['F6P'])) + "\t"
                data_list5 += str(max(0, state['G3P'])) + "\t" + str(max(0, state['Glu_in'])) + "\t"
                data_list5 += str(max(0, state['Glutamate'])) + "\t" + str(max(0, state['Glutamine_in'])) + "\t"
                data_list5 += str(max(0, state['GSH'])) + "\t" + str(max(0, state['Lac_in'])) + "\t"
                data_list5 += str(max(0, state['mAKG'])) + "\t" + str(max(0, state['MAL'])) + "\t"
                data_list5 += str(max(0, state['mASP'])) + "\t" + str(max(0, state['mCIT'])) + "\t"
                data_list5 += str(max(0, state['mFUM'])) + "\t" + str(max(0, state['mGlutamate'])) + "\t"
                data_list5 += str(max(0, state['mICIT'])) + "\t" + str(max(0, state['mMAL'])) + "\t"
                data_list5 += str(max(0, state['mOAA'])) + "\t" + str(max(0, state['mSUC'])) + "\t"
                data_list5 += str(max(0, state['NADH'])) + "\t" + str(max(0, state['NADPH'])) + "\t"
                data_list5 += str(max(0, state['O2i'])) + "\t" + str(max(0, state['OAA'])) + "\t"
                data_list5 += str(max(0, state['Glutamate'])) + "\t" + str(max(0, state['Glutamine_in'])) + "\t"
                data_list5 += str(max(0, state['PEP'])) + "\t" + str(max(0, state['PYR'])) + "\t"
                data_list5 += str(max(0, state['Pyr'])) + "\t" + str(max(0, state['R5P'])) + "\t"
                data_list5 += str(max(0, state['S7P'])) + "\t" + str(max(0, state['sixPG'])) + "\t" + str(
                    max(0, state['threePG'])) + "\t"
                data_list5 += "\n"
                self.outfile5.write(data_list5)
                # Number of cells
            TotalPCcells = 0
            TotalQCcells = 0
            TotalPScells = 0
            TotalQScells = 0
            TotalNcells = 0

            TotalPCcells_vol = 0
            TotalQCcells_vol = 0
            TotalPScells_vol = 0
            TotalQScells_vol = 0
            TotalNcells_vol = 0

            for cell in self.cellList:
                if cell.type == self.PCANCER:
                    TotalPCcells += 1
                    TotalPCcells_vol += cell.volume
                if cell.type == self.QCANCER:
                    TotalQCcells += 1
                    TotalQCcells_vol += cell.volume
                if cell.type == self.PSTEM:
                    TotalPScells += 1
                    TotalPScells_vol += cell.volume
                if cell.type == self.QSTEM:
                    TotalQScells += 1
                    TotalQScells_vol += cell.volume
                if cell.type == self.NECROTIC:
                    TotalNcells += 1
                    TotalNcells_vol += cell.volume

            Totalallcells = TotalPCcells + TotalQCcells + TotalPScells + TotalQScells
            Totalallcells_vol = TotalPCcells_vol + TotalQCcells_vol + TotalPScells_vol + TotalQScells_vol + TotalNcells_vol
            # data_list=[TotalPCcells,TotalQCcells,TotalPScells,TotalQScells,TotalNcells,Totalallcells,TotalPCcells_vol,TotalQCcells_vol,TotalPScells_vol,TotalQScells_vol,TotalNcells_vol,Totalallcells_vol]
            data_list6 = str(TotalPCcells) + "\t" + str(TotalQCcells) + "\t"
            data_list6 += str(TotalPScells) + "\t" + str(TotalQScells) + "\t"
            data_list6 += str(TotalNcells) + "\t" + str(Totalallcells) + "\t"
            data_list6 += str(TotalPCcells_vol) + "\t" + str(TotalQCcells_vol) + "\t"
            data_list6 += str(TotalPScells_vol) + "\t" + str(TotalQScells_vol) + "\t"
            data_list6 += str(TotalNcells_vol) + "\t" + str(Totalallcells_vol) + "\t"
            data_list6 += "\n"
            self.outfile6.write(data_list6)

    def finish(self):
        self.outfile1.close()
        self.outfile2.close()
        self.outfile3.close()
        self.outfile4.close()
        self.outfile5.close()
        self.outfile6.close()
        return
